# coding=utf-8
from __future__ import absolute_import

import flask
import json
import octoprint.plugin
from octoprint_authentise import helpers

class BlueprintPlugin(octoprint.plugin.BlueprintPlugin):
    @octoprint.plugin.BlueprintPlugin.route("/connect/", methods=["POST"])
    def connect(self):
        username = flask.request.json.get('username')
        password = flask.request.json.get('password')

        status_code, response_json, cookies = helpers.login(username, password, self._logger)
        if not cookies:
            return json.dumps(response_json), status_code

        status_code, response_json = helpers.create_api_token(cookies, self._logger)
        if status_code == 201:
            api_key = response_json.get('uuid')
            api_secret = response_json.get('secret')

            if helpers.claim_node(self.node_uuid, api_key, api_secret, self._logger):
                self.on_settings_save({
                    "api_key": api_key,
                    "api_secret": api_secret,
                    "claimed_by": response_json.get('user_uuid'),
                })

        return json.dumps(response_json), status_code



# coding=utf-8
from __future__ import absolute_import

import octoprint.plugin
from octoprint_authentise import blueprint, asset, settings, startup, template, update, printer

__plugin_name__ = "Authentise"

class AuthentisePlugin(
    startup.StartupPlugin,
    template.TemplatePlugin,
    settings.SettingsPlugin,
    asset.AssetPlugin,
    blueprint.BlueprintPlugin,
    update.UpdatePlugin,
    printer.AuthentisePrinter,
):
    pass

def __plugin_load__():
    global __plugin_implementation__
    __plugin_implementation__ = AuthentisePlugin()

    global __plugin_hooks__
    __plugin_hooks__ = {
        "octoprint.plugin.softwareupdate.check_config": __plugin_implementation__.get_update_information
    }



# coding=utf-8
from __future__ import absolute_import

class UpdatePlugin():
    def get_update_information(self):
        return dict(
            authentise=dict(
                displayName="Authentise Plugin",
                displayVersion=self._plugin_version,
                type="github_release",
                user="OctoPrint",
                repo="OctoPrint-Authentise",
                current=self._plugin_version,
                pip="https://github.com/Authentise/OctoPrint-Authentise/archive/{target_version}.zip"
            )
        )



# coding=utf-8
from __future__ import absolute_import
__author__ = "Scott Lemmon <scott@authentise.com> based on work by Gina Häußge"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2015 Authentise - Released under terms of the AGPLv3 License"

import threading
import logging
import urlparse

from octoprint.events import eventManager, Events
from octoprint.util import get_exception_string, RepeatedTimer, comm_helpers
import requests


AUTHENTISE_URL = 'https://print.authentise.com'
class AuthentiseMachineCom(object):
    STATE_NONE = 0
    STATE_OPEN_SERIAL = 1
    STATE_DETECT_SERIAL = 2
    STATE_DETECT_BAUDRATE = 3
    STATE_CONNECTING = 4
    STATE_OPERATIONAL = 5
    STATE_PRINTING = 6
    STATE_PAUSED = 7
    STATE_CLOSED = 8
    STATE_ERROR = 9
    STATE_CLOSED_WITH_ERROR = 10
    STATE_TRANSFERING_FILE = 11

    def __init__(self, authentise_printer_id=None, api_key=None, api_secret=None, callbackObject=None):
        self._logger = logging.getLogger(__name__)
        self._serialLogger = logging.getLogger("SERIAL")

        if callbackObject == None:
            callbackObject = MachineComPrintCallback()

        self._authentise_printer_id = authentise_printer_id
        self._command_uri_queue = comm_helpers.TypedQueue()

        self._callback = callbackObject
        self._state = self.STATE_NONE

        self._api_key = api_key
        self._api_secret = api_secret

        self._temp = {}
        self._bedTemp = None

        self._temperature_timer = None

        # print job
        self._authentise_model = None

        # monitoring thread
        self._monitoring_active = True
        self.monitoring_thread = threading.Thread(target=self._monitor_loop, name="comm._monitor")
        self.monitoring_thread.daemon = True
        self.monitoring_thread.start()

    def __del__(self):
        self.close()

    ##~~ internal state management

    def _changeState(self, newState):
        if self._state == newState:
            return

        oldState = self.getStateString()
        self._state = newState
        self._log('Changing monitoring state from \'%s\' to \'%s\'' % (oldState, self.getStateString()))
        self._callback.on_comm_state_change(newState)

    def _log(self, message):
        self._callback.on_comm_log(message)
        self._serialLogger.debug(message)

    ##~~ getters

    def getState(self):
        return self._state

    def getStateString(self):
        if self._state == self.STATE_NONE:
            return "Offline"
        if self._state == self.STATE_OPEN_SERIAL:
            return "Opening serial port"
        if self._state == self.STATE_DETECT_SERIAL:
            return "Detecting serial port"
        if self._state == self.STATE_DETECT_BAUDRATE:
            return "Detecting baudrate"
        if self._state == self.STATE_CONNECTING:
            return "Connecting"
        if self._state == self.STATE_OPERATIONAL:
            return "Operational"
        if self._state == self.STATE_PRINTING:
            return "Printing"
        if self._state == self.STATE_PAUSED:
            return "Paused"
        if self._state == self.STATE_CLOSED:
            return "Closed"
        if self._state == self.STATE_ERROR:
            return "Error: %s" % (self.getErrorString())
        if self._state == self.STATE_CLOSED_WITH_ERROR:
            return "Error: %s" % (self.getErrorString())
        if self._state == self.STATE_TRANSFERING_FILE:
            return "Transfering file to SD"
        return "?%d?" % (self._state)

    def getErrorString(self):
        return self._errorValue

    def isClosedOrError(self):
        return self._state == self.STATE_ERROR or self._state == self.STATE_CLOSED_WITH_ERROR or self._state == self.STATE_CLOSED

    def isError(self):
        return self._state == self.STATE_ERROR or self._state == self.STATE_CLOSED_WITH_ERROR

    def isOperational(self):
        return self._state == self.STATE_OPERATIONAL or self._state == self.STATE_PRINTING or self._state == self.STATE_PAUSED or self._state == self.STATE_TRANSFERING_FILE

    def isPrinting(self):
        return self._state == self.STATE_PRINTING

    def isStreaming(self):
        return self._authentise_model is not None

    def isPaused(self):
        return self._state == self.STATE_PAUSED

    def isBusy(self):
        return self.isPrinting() or self.isPaused()

    def isSdReady(self):
        return self._sdAvailable

    def getPrintProgress(self):
        #TODO: Add print progress status updates
        return self._print_progress

    def getPrintFilepos(self):
        return

    def getPrintTime(self):
        #TODO: Add print time
            return

    def getCleanedPrintTime(self):
        return

    def getTemp(self):
        return self._temp

    def getBedTemp(self):
        return self._bedTemp

    def getOffsets(self):
        return {}

    def getCurrentTool(self):
        return 0

    def getConnection(self):
        return self._authentise_printer_id

    ##~~ external interface

    def on_connected(self):
        self._temperature_timer = RepeatedTimer(lambda: comm_helpers.get_interval("temperature", default_value=4.0), self._poll_temperature, run_first=True)
        self._temperature_timer.start()

        self._changeState(self.STATE_OPERATIONAL)

        eventManager().fire(Events.CONNECTED, None)

    def close(self, isError = False):
        if self._temperature_timer is not None:
            try:
                self._temperature_timer.cancel()
            except:
                pass

        self._monitoring_active = False

        printing = self.isPrinting() or self.isPaused()

        if printing:
            eventManager().fire(Events.PRINT_FAILED, None)
        self._changeState(self.STATE_CLOSED)
        eventManager().fire(Events.DISCONNECTED)

    def setTemperatureOffset(self, offsets):
        return

    def fakeOk(self):
        return

    def sendCommand(self, cmd, cmd_type=None, processed=False):
        cmd = cmd.encode('ascii', 'replace')
        if not processed:
            cmd = comm_helpers.process_gcode_line(cmd)
            if not cmd:
                return

        if self.isPrinting() or self.isOperational():
            data = {'command': cmd}
            printer_command_url = urlparse.urljoin(AUTHENTISE_URL, 'printer/instance/{}/command/'.format(self._authentise_printer_id))
            response = requests.post(printer_command_url, json=data, auth=(self._api_key, self._api_secret))
            if not response.ok:
                self._log('Warning: Got invalid response {}: {} for {}: {}'.format(response.status_code, response.content, response.request.url, response.request.body))
                return

            self._log('Sent {} to {} with response {}: {}'.format(response.request.body, response.request.url, response.status_code, response.content))
            command_uri = response.headers['Location']
            self._command_uri_queue.put({'uri': command_uri, 'tries': 0})

    def startPrint(self):
        if not self.isOperational() or self.isPrinting():
            return

        try:
            #TODO: add logic to make print request here
            payload = {}
            response = requests.post('http://print.authentise.com/print/', json=payload, auth=(self._api_key, self._api_secret))

            self._changeState(self.STATE_PRINTING)
            eventManager().fire(Events.PRINT_STARTED, None)

        except:
            self._logger.exception("Error while trying to start printing")
            self._errorValue = get_exception_string()
            self._changeState(self.STATE_ERROR)
            eventManager().fire(Events.ERROR, {"error": self.getErrorString()})

    def selectFile(self, model_uri, sd):
        if self.isBusy():
            return

        self._authentise_model = {
                'uri': model_uri,
                'name': None,
                'snapshot': None,
                'content': None,
                'manifold': None,
                }

        response = requests.get(model_uri)
        if response.ok:
            model_data = response.json()
            self._authentise_model['name'] = model_data['name']
            self._authentise_model['status'] = model_data['status']
            self._authentise_model['snapshot'] = model_data['snapshot']
            self._authentise_model['content'] = model_data['content']
            self._authentise_model['manifold'] = model_data['analyses']['manifold']

        eventManager().fire(Events.FILE_SELECTED, {
            "file": self._authentise_model['name'],
            "filename": self._authentise_model['name'],
            "origin": self._authentise_model['uri'],
        })
        self._callback.on_comm_file_selected(model_uri, 0, False)

    def unselectFile(self):
        if self.isBusy():
            return

        self._authentise_model = None
        eventManager().fire(Events.FILE_DESELECTED)
        self._callback.on_comm_file_selected(None, None, False)

    def cancelPrint(self):
        if not self.isOperational() or self.isStreaming():
            return

        if not self._authentise_model:
            return

        self._changeState(self.STATE_OPERATIONAL)

        payload = {
            "file": self._authentise_model['name'],
            "filename": self._authentise_model['name'],
            "origin": self._authentise_model['uri'],
        }

        eventManager().fire(Events.PRINT_CANCELLED, payload)

    def setPause(self, pause):
        if self.isStreaming():
            return

        if not self._authentise_model:
            return

        payload = {
            "file": self._authentise_model['name'],
            "filename": self._authentise_model['name'],
            "origin": self._authentise_model['uri'],
        }

        if not pause and self.isPaused():
            self._changeState(self.STATE_PRINTING)
            #TODO: send resume command

            eventManager().fire(Events.PRINT_RESUMED, payload)
        elif pause and self.isPrinting():
            self._changeState(self.STATE_PAUSED)
            #TODO: send pause command

            eventManager().fire(Events.PRINT_PAUSED, payload)

    def getSdFiles(self):
        return

    def startSdFileTransfer(self, filename):
        return

    def endSdFileTransfer(self, filename):
        return

    def deleteSdFile(self, filename):
        return

    def refreshSdFiles(self):
        return

    def initSdCard(self):
        return

    def releaseSdCard(self):
        return

    ##~~ Serial monitor processing received messages

    def _monitor_loop(self):
        def _readline():
            #TODO: read the response from command uri in the command uri queue
            # if the response is older that 2 minutes, pop it from the queue    
            # if the response has a gcode response, return it and pop it from the queue
            # also should probably check if the command is a temp command
            #   and throw the updated temp event accordingly
            return ''

        self._log("Connected, starting monitor")

        while self._monitoring_active:
            try:
                line = _readline()
                if not line:
                    continue

                #TODO: if the line is a temp line:
                if 'T:' in line: #this is probably totally wrong
                    self._callback.on_comm_temperature_update(self._temp, self._bedTemp)
                self._log(line)
                self._callback.on_comm_message(line)

            except:
                self._logger.exception("Something crashed inside the serial connection loop, please report this in OctoPrint's bug tracker:")

                errorMsg = "See octoprint.log for details"
                self._log(errorMsg)
                self._errorValue = errorMsg
                self._changeState(self.STATE_ERROR)
                eventManager().fire(Events.ERROR, {"error": self.getErrorString()})
        self._log("Connection closed, closing down monitor")


    def _poll_temperature(self):
        if self.isOperational():
            self.sendCommand("M105", cmd_type="temperature_poll")




# coding=utf-8
from __future__ import absolute_import

from uuid import uuid4
import sys
import os
import subprocess
import requests
import json

# AUTHENTISE_CLIENT_PATH = 'authentise-streaming-client'
AUTHENTISE_CLIENT_PATH = '/Applications/Authentise.app/Contents/Resources/streamus-client'
AUTHENTISE_PRINT_API = 'https://print.dev-auth.com'
AUTHENTISE_USER_API = 'https://users.dev-auth.com'

def run_client(arg, logger):
    command = (
        AUTHENTISE_CLIENT_PATH, '--logging-level', 'error',
        '-c',
        '/Applications/Authentise.app/Contents/Resources/client.conf',
        arg,
    )

    try:
        return subprocess.check_output(command).strip()
        #return subprocess.check_output((AUTHENTISE_CLIENT_PATH, '--logging-level', 'error') + arg).strip()
    except Exception as e:
        logger.error("Error running client command `%s` using parameters: %s", e, args)
        return

def claim_node(node_uuid, api_key, api_secret, logger):
    if not node_uuid:
        logger.error("No node uuid available to claim")
        return False

    if not api_key:
        logger.error("No API Key available to claim node")
        return False

    if not api_secret:
        logger.error("No API secret available to claim node")
        return False

    claim_code = run_client('--connection-code', logger)
    if claim_code:
        logger.info("Got claim code: %s", claim_code)
    else:
        logger.error("Could not get a claim code from Authentise")
        return False

    url = "{}/client/claim/{}/".format(AUTHENTISE_PRINT_API, claim_code)
    response = requests.put(url, auth=(api_key, api_secret))
    logger.info("Response from - POST %s - %s - %s", url, response.status_code, response.text)

    if response.ok:
        logger.info("Claimed node: %s", node_uuid)
        return True

    logger.error("Could not use claim code %s for node %s", claim_code, node_uuid)
    return False

def login(username, password, logger):
    url = '{}/sessions/'.format(AUTHENTISE_USER_API)
    payload = {
        "username": username,
        "password": password,
    }
    response = requests.post(url, json=payload)
    logger.info("Response from - POST %s - %s - %s", url, response.status_code, response.text)

    if response.ok:
        logger.info("Successfully logged in to user service: %s", username)
        return response.status_code, json.loads(response.text), response.cookies
    elif response.status_code == 400:
        logger.warning("Failed to log in to user service for user %s", username)
    else:
        logger.error("Error logging in to user service for user %s", username)

    return response.status_code, json.loads(response.text), None

def create_api_token(cookies, logger):
    if not cookies:
        logger.warning("Cannot create api token without a valid session cookie")
        return None, None

    url = '{}/api_tokens/'.format(AUTHENTISE_USER_API)
    payload = { "name": "Octoprint Token - {}".format(str(uuid4())) }
    response = requests.post(url, json=payload, cookies=cookies)
    logger.info("Response from - POST %s - %s - %s", url, response.status_code, response.text)

    if response.ok:
        logger.info("Successfully created api token: %s", response.text)
    elif response.status_code == 400:
        logger.warning("Failed to create api token for user %s", username)
    else:
        logger.error("Error creating api token for user")

    return response.status_code, json.loads(response.text)



# coding=utf-8
from __future__ import absolute_import

import octoprint.plugin
from octoprint_authentise import helpers

class StartupPlugin(octoprint.plugin.StartupPlugin):
    def on_after_startup(self):
        self.node_version = helpers.run_client('--version', self._logger)
        if self.node_version:
            self._logger.info("Found node version: %s", self.node_version)
        else:
            self._logger.warning("Could not find node version")

        self.node_uuid = helpers.run_client('--node-uuid', self._logger)
        if self.node_uuid:
            self._logger.info("Found node uuid: %s", self.node_uuid)
        else:
            self._logger.warning("Could not find node uuid")



# coding=utf-8
from __future__ import absolute_import

import octoprint.plugin

class SettingsPlugin(octoprint.plugin.SettingsPlugin):
    def get_settings_defaults(self):
        return dict(
            api_key=None,
            api_secret=None,
            claimed_by=None,
        )



# coding=utf-8
from setuptools import setup
import contextlib
import imp
import os
import re
import subprocess
try:
    import octoprint_setuptools
except:
    print("Could not import OctoPrint's setuptools, are you sure you are running that under "
        "the same python installation that OctoPrint is installed under?")
    import sys
    sys.exit(-1)

########################################################################################################################
### Do not forget to adjust the following variables to your own plugin.

# The plugin's identifier, has to be unique
plugin_identifier = "authentise"

# The plugin's python package, should be "octoprint_<plugin identifier>", has to be unique
plugin_package = "octoprint_authentise"

# The plugin's human readable name. Can be overwritten within OctoPrint's internal data via __plugin_name__ in the
# plugin module
plugin_name = "OctoPrint-Authentise"

# The plugin's description. Can be overwritten within OctoPrint's internal data via __plugin_description__ in the plugin
# module
plugin_description = """Authenitse Octoprint Integration"""

# The plugin's author. Can be overwritten within OctoPrint's internal data via __plugin_author__ in the plugin module
plugin_author = "Authentise, Inc."

# The plugin's author's mail address.
plugin_author_email = "engineering@authentise.com"

# The plugin's homepage URL. Can be overwritten within OctoPrint's internal data via __plugin_url__ in the plugin module
plugin_url = "https://github.com/authentise/OctoPrint-Authentise"

# The plugin's license. Can be overwritten within OctoPrint's internal data via __plugin_license__ in the plugin module
plugin_license = "AGPLv3"

# Any additional requirements besides OctoPrint should be listed here
plugin_requires = []

### --------------------------------------------------------------------------------------------------------------------
### More advanced options that you usually shouldn't have to touch follow after this point
### --------------------------------------------------------------------------------------------------------------------

# Additional package data to install for this plugin. The subfolders "templates", "static" and "translations" will
# already be installed automatically if they exist.
plugin_additional_data = []

# Any additional python packages you need to install with your plugin that are not contains in <plugin_package>.*
plugin_addtional_packages = []

# Any python packages within <plugin_package>.* you do NOT want to install with your plugin
plugin_ignored_packages = []

# Additional parameters for the call to setuptools.setup. If your plugin wants to register additional entry points,
# define dependency links or other things like that, this is the place to go. Will be merged recursively with the
# default setup parameters as provided by octoprint_setuptools.create_plugin_setup_parameters using
# octoprint.util.dict_merge.
#
# Example:
#     plugin_requires = ["someDependency==dev"]
#     additional_setup_parameters = {"dependency_links": ["https://github.com/someUser/someRepo/archive/master.zip#egg=someDependency-dev"]}
additional_setup_parameters = {}

########################################################################################################################

VERSION_FILE = 'octoprint_authentise/version.py'

def _get_output_or_none(args):
    def _check_output(*popenargs, **kwargs):
        r"""Run command with arguments and return its output as a byte string.
        Backported from Python 2.7 as it's implemented as pure python on stdlib.
        >>> check_output(['/usr/bin/python', '--version'])
        Python 2.6.2
        """
        process = subprocess.Popen(stdout=subprocess.PIPE, *popenargs, **kwargs)
        output, unused_err = process.communicate()
        retcode = process.poll()
        if retcode:
            cmd = kwargs.get("args")
            if cmd is None:
                cmd = popenargs[0]
            error = subprocess.CalledProcessError(retcode, cmd)
            error.output = output
            raise error
        return output

    try:
        return _check_output(args).decode('utf-8').strip()
    except subprocess.CalledProcessError:
        return None

def _get_git_description():
    return _get_output_or_none(['git', 'describe'])

def _get_git_branches_for_this_commit():
    branches = _get_output_or_none(['git', 'branch', '-r', '--contains', 'HEAD'])
    split = branches.split('\n') if branches else []
    return [branch.strip() for branch in split]

def _is_on_releasable_branch(branches):
    return any(['origin/master' == branch or branch.startswith('origin/hotfix') for branch in branches])

def _git_to_version(git):
    match = re.match(r'(?P<tag>[\d\.]+)-(?P<offset>[\d]+)-(?P<sha>\w{8})', git)
    if not match:
        version = git
    else:
        version = "{tag}.post0.dev{offset}".format(**match.groupdict())
    print("Calculated {0} version '{1}' from git description '{2}'".format(plugin_package, version, git))
    return version

@contextlib.contextmanager
def write_version():
    git_description = _get_git_description()
    git_branches = _get_git_branches_for_this_commit()
    version = _git_to_version(git_description) if git_description else None
    if git_branches and not _is_on_releasable_branch(git_branches):
        print("Forcing version to 0.0.1 because this commit is on branches {0} and not a whitelisted branch".format(git_branches))
        version = '0.0.1'
    if version:
        with open(VERSION_FILE, 'r') as version_file:
            old_contents = version_file.read()
        with open(VERSION_FILE, 'w') as version_file:
            version_file.write('VERSION = "{0}"\n'.format(version))
    yield
    if version:
        with open(VERSION_FILE, 'w') as version_file:
            version_file.write(old_contents)

def get_version():
    basedir = os.path.abspath(os.path.dirname(__file__))
    version = imp.load_source('version', os.path.join(basedir, plugin_package, 'version.py'))
    return version.VERSION

with write_version():
    setup_parameters = octoprint_setuptools.create_plugin_setup_parameters(
            identifier=plugin_identifier,
            package=plugin_package,
            name=plugin_name,
            version=get_version(),
            description=plugin_description,
            author=plugin_author,
            mail=plugin_author_email,
            url=plugin_url,
            license=plugin_license,
            requires=plugin_requires,
            additional_packages=plugin_addtional_packages,
            ignored_packages=plugin_ignored_packages,
            additional_data=plugin_additional_data
    )

    if len(additional_setup_parameters):
        from octoprint.util import dict_merge
        setup_parameters = dict_merge(setup_parameters, additional_setup_parameters)

    setup(**setup_parameters)



# coding=utf-8
"""
This module holds the Authentise web based printing service specific implementation of the `PrinterPlugin` and its helpers.
"""
from __future__ import absolute_import
__author__ = "Scott Lemmon <scott@authentise.com>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2015 Authentise - Released under terms of the AGPLv3 License"

import copy
import logging
import threading
import time
import re

from . import comm

import octoprint.plugin

from octoprint.events import eventManager, Events
from octoprint.filemanager import FileDestinations
from octoprint.printer import PrinterCallback
from octoprint.printer.estimation import TimeEstimationHelper
from octoprint.settings import settings
from octoprint.util import InvariantContainer


class AuthentisePrinter(octoprint.plugin.PrinterPlugin,
                        octoprint.plugin.MachineComPrintCallbackPlugin):
    """
    Authentise web based printing service specific implementation of the `PrinterPlugin`. Manages the communication layer object and registers itself with it as a callback to react to changes on the communication layer.
    """

    valid_axes = ("x", "y", "z", "e")

    valid_tool_regex = re.compile("^(tool\d+)$")

    valid_heater_regex = re.compile("^(tool\d+|bed)$")

    def startup(self, fileManager, analysisQueue, printerProfileManager):
        from collections import deque

        self._logger = logging.getLogger(__name__)

        self._analysisQueue = analysisQueue
        self._fileManager = fileManager
        self._printerProfileManager = printerProfileManager

        # state
        # TODO do we really need to hold the temperature here?
        self._temp = None
        self._bedTemp = None
        self._targetTemp = None
        self._targetBedTemp = None
        self._temps = TemperatureHistory(cutoff=settings().getInt(["temperature", "cutoff"])*60)
        self._tempBacklog = []

        self._latestMessage = None
        self._messages = deque([], 300)
        self._messageBacklog = []

        self._latestLog = None
        self._log = deque([], 300)
        self._logBacklog = []

        self._state = None

        self._currentZ = None

        self._progress = None
        self._printTime = None
        self._printTimeLeft = None

        self._printAfterSelect = False

        # sd handling
        self._sdPrinting = False
        self._sdStreaming = False
        self._sdFilelistAvailable = threading.Event()
        self._streamingFinishedCallback = None

        self._selectedFile = None
        self._timeEstimationData = None

        # comm
        self._comm = None

        # callbacks
        self._callbacks = []

        # progress plugins
        self._lastProgressReport = None
        self._progressPlugins = octoprint.plugin.plugin_manager().get_implementations(octoprint.plugin.ProgressPlugin)

        self._stateMonitor = StateMonitor(
            interval=0.5,
            on_update=self._sendCurrentDataCallbacks,
            on_add_temperature=self._sendAddTemperatureCallbacks,
            on_add_log=self._sendAddLogCallbacks,
            on_add_message=self._sendAddMessageCallbacks
        )
        self._stateMonitor.reset(
            state={"text": self.get_state_string(), "flags": self._getStateFlags()},
            job_data={
                "file": {
                    "name": None,
                    "size": None,
                    "origin": None,
                    "date": None
                },
                "estimatedPrintTime": None,
                "lastPrintTime": None,
                "filament": {
                    "length": None,
                    "volume": None
                }
            },
            progress={"completion": None, "filepos": None, "printTime": None, "printTimeLeft": None},
            current_z=None
        )

        eventManager().subscribe(Events.METADATA_ANALYSIS_FINISHED, self._on_event_MetadataAnalysisFinished)
        eventManager().subscribe(Events.METADATA_STATISTICS_UPDATED, self._on_event_MetadataStatisticsUpdated)

    #~~ handling of PrinterCallbacks

    def register_callback(self, callback):
        if not isinstance(callback, PrinterCallback):
            self._logger.warn("Registering an object as printer callback which doesn't implement the PrinterCallback interface")

        self._callbacks.append(callback)
        self._sendInitialStateUpdate(callback)

    def unregister_callback(self, callback):
        if callback in self._callbacks:
            self._callbacks.remove(callback)

    def _sendAddTemperatureCallbacks(self, data):
        for callback in self._callbacks:
            try: callback.on_printer_add_temperature(data)
            except: self._logger.exception("Exception while adding temperature data point")

    def _sendAddLogCallbacks(self, data):
        for callback in self._callbacks:
            try: callback.on_printer_add_log(data)
            except: self._logger.exception("Exception while adding communication log entry")

    def _sendAddMessageCallbacks(self, data):
        for callback in self._callbacks:
            try: callback.on_printer_add_message(data)
            except: self._logger.exception("Exception while adding printer message")

    def _sendCurrentDataCallbacks(self, data):
        for callback in self._callbacks:
            try: callback.on_printer_send_current_data(copy.deepcopy(data))
            except: self._logger.exception("Exception while pushing current data")

    #~~ callback from metadata analysis event

    def _on_event_MetadataAnalysisFinished(self, event, data):
        if self._selectedFile:
            self._setJobData(self._selectedFile["filename"])

    def _on_event_MetadataStatisticsUpdated(self, event, data):
        self._setJobData(self._selectedFile["filename"])

    #~~ progress plugin reporting

    def _reportPrintProgressToPlugins(self, progress):
        if not progress or not self._selectedFile or not "sd" in self._selectedFile or not "filename" in self._selectedFile:
            return

        storage = "sdcard" if self._selectedFile["sd"] else "local"
        filename = self._selectedFile["filename"]

        def call_plugins(storage, filename, progress):
            for plugin in self._progressPlugins:
                try:
                    plugin.on_print_progress(storage, filename, progress)
                except:
                    self._logger.exception("Exception while sending print progress to plugin %s" % plugin._identifier)

        thread = threading.Thread(target=call_plugins, args=(storage, filename, progress))
        thread.daemon = False
        thread.start()

    #~~ PrinterPlugin implementation

    def connect(self, port=None, baudrate=None, profile=None):
        #TODO: This will be gotten from a request eventually, but currently hardcoded for testing
        authentise_printer_id = 1901
        plugin_settings = settings().get(['plugins', 'authentise'])
        api_key = plugin_settings['api_key']
        api_secret = plugin_settings['api_secret']

        if self._comm is not None:
            self._comm.close()
        self._printerProfileManager.select(profile)
        self._comm = comm.AuthentiseMachineCom(authentise_printer_id=authentise_printer_id, api_key=api_key, api_secret=api_secret, callbackObject=self)
        self._comm.on_connected()

    def disconnect(self):
        if self._comm is not None:
            self._comm.close()
        self._comm = None
        self._printerProfileManager.deselect()
        eventManager().fire(Events.DISCONNECTED)

    def get_transport(self):
        return

    def fake_ack(self):
        if self._comm is None:
            return

        self._comm.fakeOk()

    def commands(self, commands):
        if self._comm is None:
            return

        if not isinstance(commands, (list, tuple)):
            commands = [commands]

        for command in commands:
            self._comm.sendCommand(command)

    def script(self, name, context=None):
        return

    def jog(self, axis, amount):
        if not isinstance(axis, (str, unicode)):
            raise ValueError("axis must be a string: {axis}".format(axis=axis))

        axis = axis.lower()
        if not axis in self.valid_axes:
            raise ValueError("axis must be any of {axes}: {axis}".format(axes=", ".join(self.valid_axes), axis=axis))
        if not isinstance(amount, (int, long, float)):
            raise ValueError("amount must be a valid number: {amount}".format(amount=amount))

        printer_profile = self._printerProfileManager.get_current_or_default()
        movement_speed = printer_profile["axes"][axis]["speed"]
        self.commands(["G91", "G1 %s%.4f F%d" % (axis.upper(), amount, movement_speed), "G90"])

    def home(self, axes):
        if not isinstance(axes, (list, tuple)):
            if isinstance(axes, (str, unicode)):
                axes = [axes]
            else:
                raise ValueError("axes is neither a list nor a string: {axes}".format(axes=axes))

        validated_axes = filter(lambda x: x in self.valid_axes, map(lambda x: x.lower(), axes))
        if len(axes) != len(validated_axes):
            raise ValueError("axes contains invalid axes: {axes}".format(axes=axes))

        self.commands(["G91", "G28 %s" % " ".join(map(lambda x: "%s0" % x.upper(), validated_axes)), "G90"])

    def extrude(self, amount):
        if not isinstance(amount, (int, long, float)):
            raise ValueError("amount must be a valid number: {amount}".format(amount=amount))

        printer_profile = self._printerProfileManager.get_current_or_default()
        extrusion_speed = printer_profile["axes"]["e"]["speed"]
        self.commands(["G91", "G1 E%s F%d" % (amount, extrusion_speed), "G90"])

    def change_tool(self, tool):
        if not self.valid_tool_regex.match(tool):
            raise ValueError("tool must match \"tool[0-9]+\": {tool}".format(tool=tool))

        tool_num = int(tool[len("tool"):])
        self.commands("T%d" % tool_num)

    def set_temperature(self, heater, value):
        if not self.valid_heater_regex.match(heater):
            raise ValueError("heater must match \"tool[0-9]+\" or \"bed\": {heater}".format(type=heater))

        if not isinstance(value, (int, long, float)) or value < 0:
            raise ValueError("value must be a valid number >= 0: {value}".format(value=value))

        if heater.startswith("tool"):
            printer_profile = self._printerProfileManager.get_current_or_default()
            extruder_count = printer_profile["extruder"]["count"]
            if extruder_count > 1:
                toolNum = int(heater[len("tool"):])
                self.commands("M104 T%d S%f" % (toolNum, value))
            else:
                self.commands("M104 S%f" % value)

        elif heater == "bed":
            self.commands("M140 S%f" % value)

    def set_temperature_offset(self, offsets=None):
        if offsets is None:
            offsets = dict()

        if not isinstance(offsets, dict):
            raise ValueError("offsets must be a dict")

        validated_keys = filter(lambda x: self.valid_heater_regex.match(x), offsets.keys())
        validated_values = filter(lambda x: isinstance(x, (int, long, float)), offsets.values())

        if len(validated_keys) != len(offsets):
            raise ValueError("offsets contains invalid keys: {offsets}".format(offsets=offsets))
        if len(validated_values) != len(offsets):
            raise ValueError("offsets contains invalid values: {offsets}".format(offsets=offsets))

        if self._comm is None:
            return

        self._comm.setTemperatureOffset(offsets)
        self._stateMonitor.set_temp_offsets(offsets)

    def _convert_rate_value(self, factor, min=0, max=200):
        if not isinstance(factor, (int, float, long)):
            raise ValueError("factor is not a number")

        if isinstance(factor, float):
            factor = int(factor * 100.0)

        if factor < min or factor > max:
            raise ValueError("factor must be a value between %f and %f" % (min, max))

        return factor

    def feed_rate(self, factor):
        factor = self._convert_rate_value(factor, min=50, max=200)
        self.commands("M220 S%d" % factor)

    def flow_rate(self, factor):
        factor = self._convert_rate_value(factor, min=75, max=125)
        self.commands("M221 S%d" % factor)

    def select_file(self, path, sd, printAfterSelect=False):
        if self._comm is None or (self._comm.isBusy() or self._comm.isStreaming()):
            self._logger.info("Cannot load file: printer not connected or currently busy")
            return

        self._printAfterSelect = printAfterSelect
        self._comm.selectFile("/" + path if sd else path, sd)
        self._setProgressData(0, None, None, None)
        self._setCurrentZ(None)

    def unselect_file(self):
        if self._comm is not None and (self._comm.isBusy() or self._comm.isStreaming()):
            return

        self._comm.unselectFile()
        self._setProgressData(0, None, None, None)
        self._setCurrentZ(None)

    def start_print(self):
        if self._comm is None or not self._comm.isOperational() or self._comm.isPrinting():
            return
        if self._selectedFile is None:
            return

        rolling_window = None
        threshold = None
        countdown = None
        if self._selectedFile["sd"]:
            # we are interesting in a rolling window of roughly the last 15s, so the number of entries has to be derived
            # by that divided by the sd status polling interval
            rolling_window = 15 / settings().get(["serial", "timeout", "sdStatus"])

            # we are happy if the average of the estimates stays within 60s of the prior one
            threshold = 60

            # we are happy when one rolling window has been stable
            countdown = rolling_window
        self._timeEstimationData = TimeEstimationHelper(rolling_window=rolling_window, threshold=threshold, countdown=countdown)

        self._lastProgressReport = None
        self._setCurrentZ(None)
        self._comm.startPrint()

    def toggle_pause_print(self):
        if self._comm is None:
            return

        self._comm.setPause(not self._comm.isPaused())

    def cancel_print(self):
        if self._comm is None:
            return

        self._comm.cancelPrint()

        # reset progress, height, print time
        self._setCurrentZ(None)
        self._setProgressData(None, None, None, None)

        # mark print as failure
        if self._selectedFile is not None:
            self._fileManager.log_print(FileDestinations.SDCARD if self._selectedFile["sd"] else FileDestinations.LOCAL, self._selectedFile["filename"], time.time(), self._comm.getPrintTime(), False, self._printerProfileManager.get_current_or_default()["id"])
            payload = {
                "file": self._selectedFile["filename"],
                "origin": FileDestinations.LOCAL
            }
            if self._selectedFile["sd"]:
                payload["origin"] = FileDestinations.SDCARD
            eventManager().fire(Events.PRINT_FAILED, payload)

    def get_state_string(self):
        if self._comm is None:
            return "Offline"
        else:
            return self._comm.getStateString()

    def get_current_data(self):
        return self._stateMonitor.get_current_data()

    def get_current_job(self):
        currentData = self._stateMonitor.get_current_data()
        return currentData["job"]

    def get_current_temperatures(self):
        if self._comm is not None:
            offsets = self._comm.getOffsets()
        else:
            offsets = dict()

        result = {}
        if self._temp is not None:
            for tool in self._temp.keys():
                result["tool%d" % tool] = {
                    "actual": self._temp[tool][0],
                    "target": self._temp[tool][1],
                    "offset": offsets[tool] if tool in offsets and offsets[tool] is not None else 0
                }
        if self._bedTemp is not None:
            result["bed"] = {
                "actual": self._bedTemp[0],
                "target": self._bedTemp[1],
                "offset": offsets["bed"] if "bed" in offsets and offsets["bed"] is not None else 0
            }

        return result

    def get_temperature_history(self):
        return self._temps

    def get_current_connection(self):
        if self._comm is None:
            return "Closed", None, None, None

        authentise_printer_id  = self._comm.getConnection()
        printer_profile = self._printerProfileManager.get_current_or_default()
        return self._comm.getStateString(), None, None, printer_profile

    def is_closed_or_error(self):
        return self._comm is None or self._comm.isClosedOrError()

    def is_operational(self):
        return self._comm is not None and self._comm.isOperational()

    def is_printing(self):
        return self._comm is not None and self._comm.isPrinting()

    def is_paused(self):
        return self._comm is not None and self._comm.isPaused()

    def is_error(self):
        return self._comm is not None and self._comm.isError()

    def is_ready(self):
        return self.is_operational() and not self._comm.isStreaming()

    def is_sd_ready(self):
        return False

    #~~ sd file handling

    def get_sd_files(self):
        return

    def add_sd_file(self, filename, absolutePath, streamingFinishedCallback):
        return

    def delete_sd_file(self, filename):
        return

    def init_sd_card(self):
        return

    def release_sd_card(self):
        return

    def refresh_sd_files(self, blocking=False):
        return

    #~~ state monitoring

    def _setCurrentZ(self, currentZ):
        self._currentZ = currentZ
        self._stateMonitor.set_current_z(self._currentZ)

    def _setState(self, state):
        self._state = state
        self._stateMonitor.set_state({"text": self.get_state_string(), "flags": self._getStateFlags()})

    def _addLog(self, log):
        self._log.append(log)
        self._stateMonitor.add_log(log)

    def _addMessage(self, message):
        self._messages.append(message)
        self._stateMonitor.add_message(message)

    def _estimateTotalPrintTime(self, progress, printTime):
        if not progress or not printTime or not self._timeEstimationData:
            return None

        else:
            newEstimate = printTime / progress
            self._timeEstimationData.update(newEstimate)

            result = None
            if self._timeEstimationData.is_stable():
                result = self._timeEstimationData.average_total_rolling

            return result

    def _setProgressData(self, progress, filepos, printTime, cleanedPrintTime):
        estimatedTotalPrintTime = self._estimateTotalPrintTime(progress, cleanedPrintTime)
        totalPrintTime = estimatedTotalPrintTime

        if self._selectedFile and "estimatedPrintTime" in self._selectedFile and self._selectedFile["estimatedPrintTime"]:
            statisticalTotalPrintTime = self._selectedFile["estimatedPrintTime"]
            if progress and cleanedPrintTime:
                if estimatedTotalPrintTime is None:
                    totalPrintTime = statisticalTotalPrintTime
                else:
                    if progress < 0.5:
                        sub_progress = progress * 2
                    else:
                        sub_progress = 1.0
                    totalPrintTime = (1 - sub_progress) * statisticalTotalPrintTime + sub_progress * estimatedTotalPrintTime

        self._progress = progress
        self._printTime = printTime
        self._printTimeLeft = totalPrintTime - cleanedPrintTime if (totalPrintTime is not None and cleanedPrintTime is not None) else None

        self._stateMonitor.set_progress({
            "completion": self._progress * 100 if self._progress is not None else None,
            "filepos": filepos,
            "printTime": int(self._printTime) if self._printTime is not None else None,
            "printTimeLeft": int(self._printTimeLeft) if self._printTimeLeft is not None else None
        })

        if progress:
            progress_int = int(progress * 100)
            if self._lastProgressReport != progress_int:
                self._lastProgressReport = progress_int
                self._reportPrintProgressToPlugins(progress_int)


    def _addTemperatureData(self, temp, bedTemp):
        currentTimeUtc = int(time.time())

        data = {
            "time": currentTimeUtc
        }
        for tool in temp.keys():
            data["tool%d" % tool] = {
                "actual": temp[tool][0],
                "target": temp[tool][1]
            }
        if bedTemp is not None and isinstance(bedTemp, tuple):
            data["bed"] = {
                "actual": bedTemp[0],
                "target": bedTemp[1]
            }

        self._temps.append(data)

        self._temp = temp
        self._bedTemp = bedTemp

        self._stateMonitor.add_temperature(data)

    def _setJobData(self, authentise_model_uri):
        self._selectedFile = authentise_model_uri

        self._stateMonitor.set_job_data({
            "file": {
                "name": None,
                "origin": None,
                "size": None,
                "date": None
            },
            "estimatedPrintTime": None,
            "averagePrintTime": None,
            "lastPrintTime": None,
            "filament": None,
        })

    def _sendInitialStateUpdate(self, callback):
        try:
            data = self._stateMonitor.get_current_data()
            data.update({
                "temps": list(self._temps),
                "logs": list(self._log),
                "messages": list(self._messages)
            })
            callback.on_printer_send_initial_data(data)
        except Exception, err:
            import sys
            sys.stderr.write("ERROR: %s\n" % str(err))
            pass

    def _getStateFlags(self):
        return {
            "operational": self.is_operational(),
            "printing": self.is_printing(),
            "closedOrError": self.is_closed_or_error(),
            "error": self.is_error(),
            "paused": self.is_paused(),
            "ready": self.is_ready(),
            "sdReady": self.is_sd_ready()
        }

    #~~ comm.MachineComPrintCallback implementation

    def on_comm_log(self, message):
        self._addLog(message)

    def on_comm_temperature_update(self, temp, bedTemp):
        self._addTemperatureData(temp, bedTemp)

    def on_comm_state_change(self, state):
        oldState = self._state

        # forward relevant state changes to gcode manager
        if oldState == comm.AuthentiseMachineCom.STATE_PRINTING:
            if self._selectedFile is not None:
                if state == comm.AuthentiseMachineCom.STATE_CLOSED or state == comm.AuthentiseMachineCom.STATE_ERROR or state == comm.AuthentiseMachineCom.STATE_CLOSED_WITH_ERROR:
                    self._fileManager.log_print(FileDestinations.SDCARD if self._selectedFile["sd"] else FileDestinations.LOCAL, self._selectedFile["filename"], time.time(), self._comm.getPrintTime(), False, self._printerProfileManager.get_current_or_default()["id"])
            self._analysisQueue.resume() # printing done, put those cpu cycles to good use
        elif state == comm.AuthentiseMachineCom.STATE_PRINTING:
            self._analysisQueue.pause() # do not analyse files while printing
        elif state == comm.AuthentiseMachineCom.STATE_CLOSED or state == comm.AuthentiseMachineCom.STATE_CLOSED_WITH_ERROR:
            if self._comm is not None:
                self._comm = None

            self._setProgressData(0, None, None, None)
            self._setCurrentZ(None)
            self._setJobData(None)

        self._setState(state)

    def on_comm_message(self, message):
        self._addMessage(message)

    def on_comm_progress(self):
        self._setProgressData(self._comm.getPrintProgress(), self._comm.getPrintFilepos(), self._comm.getPrintTime(), self._comm.getCleanedPrintTime())

    def on_comm_z_change(self, newZ):
        oldZ = self._currentZ
        if newZ != oldZ:
            # we have to react to all z-changes, even those that might "go backward" due to a slicer's retraction or
            # anti-backlash-routines. Event subscribes should individually take care to filter out "wrong" z-changes
            eventManager().fire(Events.Z_CHANGE, {"new": newZ, "old": oldZ})

        self._setCurrentZ(newZ)

    def on_comm_sd_state_change(self, sdReady):
        self._stateMonitor.set_state({"text": self.get_state_string(), "flags": self._getStateFlags()})

    def on_comm_sd_files(self, files):
        eventManager().fire(Events.UPDATED_FILES, {"type": "gcode"})
        self._sdFilelistAvailable.set()

    def on_comm_file_selected(self, filename, filesize, sd):
        self._setJobData(filename)
        self._stateMonitor.set_state({"text": self.get_state_string(), "flags": self._getStateFlags()})

        if self._printAfterSelect:
            self.start_print()

    def on_comm_print_job_done(self):
        self._fileManager.log_print(FileDestinations.SDCARD if self._selectedFile["sd"] else FileDestinations.LOCAL, self._selectedFile["filename"], time.time(), self._comm.getPrintTime(), True, self._printerProfileManager.get_current_or_default()["id"])
        self._setProgressData(1.0, self._selectedFile["filesize"], self._comm.getPrintTime(), 0)
        self._stateMonitor.set_state({"text": self.get_state_string(), "flags": self._getStateFlags()})

    def on_comm_file_transfer_started(self, filename, filesize):
        self._sdStreaming = True

        self._setJobData(filename)
        self._setProgressData(0.0, 0, 0, None)
        self._stateMonitor.set_state({"text": self.get_state_string(), "flags": self._getStateFlags()})

    def on_comm_file_transfer_done(self, filename):
        self._sdStreaming = False

        if self._streamingFinishedCallback is not None:
            # in case of SD files, both filename and absolutePath are the same, so we set the (remote) filename for
            # both parameters
            self._streamingFinishedCallback(filename, filename, FileDestinations.SDCARD)

        self._setCurrentZ(None)
        self._setJobData(None)
        self._setProgressData(None, None, None, None)
        self._stateMonitor.set_state({"text": self.get_state_string(), "flags": self._getStateFlags()})

    def on_comm_force_disconnect(self):
        self.disconnect()


class StateMonitor(object):
    def __init__(self, interval=0.5, on_update=None, on_add_temperature=None, on_add_log=None, on_add_message=None):
        self._interval = interval
        self._update_callback = on_update
        self._on_add_temperature = on_add_temperature
        self._on_add_log = on_add_log
        self._on_add_message = on_add_message

        self._state = None
        self._job_data = None
        self._gcode_data = None
        self._sd_upload_data = None
        self._current_z = None
        self._progress = None

        self._offsets = {}

        self._change_event = threading.Event()
        self._state_lock = threading.Lock()

        self._last_update = time.time()
        self._worker = threading.Thread(target=self._work)
        self._worker.daemon = True
        self._worker.start()

    def reset(self, state=None, job_data=None, progress=None, current_z=None):
        self.set_state(state)
        self.set_job_data(job_data)
        self.set_progress(progress)
        self.set_current_z(current_z)

    def add_temperature(self, temperature):
        self._on_add_temperature(temperature)
        self._change_event.set()

    def add_log(self, log):
        self._on_add_log(log)
        self._change_event.set()

    def add_message(self, message):
        self._on_add_message(message)
        self._change_event.set()

    def set_current_z(self, current_z):
        self._current_z = current_z
        self._change_event.set()

    def set_state(self, state):
        with self._state_lock:
            self._state = state
            self._change_event.set()

    def set_job_data(self, job_data):
        self._job_data = job_data
        self._change_event.set()

    def set_progress(self, progress):
        self._progress = progress
        self._change_event.set()

    def set_temp_offsets(self, offsets):
        self._offsets = offsets
        self._change_event.set()

    def _work(self):
        while True:
            self._change_event.wait()

            with self._state_lock:
                now = time.time()
                delta = now - self._last_update
                additional_wait_time = self._interval - delta
                if additional_wait_time > 0:
                    time.sleep(additional_wait_time)

                data = self.get_current_data()
                self._update_callback(data)
                self._last_update = time.time()
                self._change_event.clear()

    def get_current_data(self):
        return {
            "state": self._state,
            "job": self._job_data,
            "currentZ": self._current_z,
            "progress": self._progress,
            "offsets": self._offsets
        }


class TemperatureHistory(InvariantContainer):
    def __init__(self, cutoff=30 * 60):

        def temperature_invariant(data):
            data.sort(key=lambda x: x["time"])
            now = int(time.time())
            return [item for item in data if item["time"] >= now - cutoff]

        InvariantContainer.__init__(self, guarantee_invariant=temperature_invariant)




# coding=utf-8
from __future__ import absolute_import

import octoprint.plugin

class TemplatePlugin(octoprint.plugin.TemplatePlugin):
    def get_template_vars(self):
        return dict(
            node_uuid=(self.node_uuid or 'Unknown'),
            node_version=(self.node_version or 'Unknown'),
            version=self._plugin_version,
        )

    def get_template_configs(self):
        return [
            dict(type="settings", custom_bindings=False),
            dict(type="navbar", classes=['dropdown']),
        ]



# coding=utf-8
from __future__ import absolute_import

import octoprint.plugin

class AssetPlugin(octoprint.plugin.AssetPlugin):
    def get_assets(self):
        return dict(
            js=["js/authentise.js"],
            less=["less/authentise.less"],
        )
